#include "socket_vol.h"


typedef struct socket_vol_server_info_t {
    struct sockaddr_in server_sockaddr;
    int sock;
    uint64_t fapl_ref;
    uint64_t obj_ref;
} socket_vol_server_info_t;

typedef struct socket_vol_info_t {
    socket_vol_server_info_t *server_info;
} socket_vol_info_t;

typedef struct socket_vol_t {
    int64_t obj_id;
    socket_vol_server_info_t *server_info;
} socket_vol_t;


static int socket_vol_init(void);
static int socket_vol_connect(socket_vol_server_info_t *server_info);
static int socket_vol_obj_free(socket_vol_t *obj);
static void *socket_vol_fapl_copy(const void *_old_info);
static herr_t socket_vol_fapl_free(void *info);
static void *socket_vol_file_create(const char *name, unsigned flags,
    hid_t fcpl_id, hid_t fapl_id, hid_t req);
static void *socket_vol_file_open(const char *name, unsigned flags,
    hid_t fapl_id, hid_t req);
static herr_t socket_vol_file_close(void *_file, hid_t req);
static void *socket_vol_group_create(void *_obj, H5VL_loc_params_t loc_params,
    const char *name, hid_t gcpl_id, hid_t gapl_id, hid_t req);
static void *socket_vol_group_open(void *_obj, H5VL_loc_params_t loc_params,
    const char *name, hid_t gapl_id, hid_t req);
static herr_t socket_vol_group_close(void *_grp, hid_t req);
static void *socket_vol_dataset_create(void *_obj, H5VL_loc_params_t loc_params,
    const char *name, hid_t dcpl_id, hid_t dapl_id, hid_t req);
static void *socket_vol_dataset_open(void *_obj, H5VL_loc_params_t loc_params,
    const char *name, hid_t dapl_id, hid_t req);
static herr_t socket_vol_dataset_write(void *_dset, hid_t mem_type_id,
    hid_t mem_space_id, hid_t file_space_id, hid_t dxpl_id, const void *app_buf,
    hid_t req);
static herr_t socket_vol_dataset_read(void *_dset, hid_t mem_type_id,
    hid_t mem_space_id, hid_t file_space_id, hid_t dxpl_id, void *app_buf,
    hid_t req);
static herr_t socket_vol_dataset_close(void *_grp, hid_t req);


static int socket_vol_init_g = 0;

static hid_t socket_vol_id_g = -1;

static const H5VL_class_t socket_vol_g = {
    DUMMY,                                      /* class */
    "socket",                                   /* name */
    NULL,                                       /* initialize */
    NULL,                                       /* terminate */
    sizeof(socket_vol_info_t),                  /* fapl_size */
    socket_vol_fapl_copy,                       /* fapl_copy */
    socket_vol_fapl_free,                       /* fapl_free */
    {                                           /* attribute_cls */
        NULL, //socket_vol_attr_create,                /* create */
        NULL, //socket_vol_attr_open,                  /* open */
        NULL, //socket_vol_attr_read,                  /* read */
        NULL, //socket_vol_attr_write,                 /* write */
        NULL, //socket_vol_attr_get,                   /* get */
        NULL, //socket_vol_attr_remove,                /* remove */
        NULL, //socket_vol_attr_close                  /* close */
    },
    {                                           /* datatype_cls */
        NULL, //socket_vol_datatype_commit,            /* commit */
        NULL, //socket_vol_datatype_open,              /* open */
        NULL, //socket_vol_datatype_get_binary,        /* get_size */
        NULL, //socket_vol_datatype_close              /* close */
    },
    {                                           /* dataset_cls */
        socket_vol_dataset_create,                     /* create */
        socket_vol_dataset_open,                       /* open */
        socket_vol_dataset_read,                       /* read */
        socket_vol_dataset_write,                      /* write */
        NULL, //socket_vol_dataset_set_extent,         /* set extent */
        NULL, //socket_vol_dataset_get,                /* get */
        socket_vol_dataset_close                       /* close */
    },
    {                                           /* file_cls */
        socket_vol_file_create,                        /* create */
        socket_vol_file_open,                          /* open */
        NULL, //socket_vol_file_flush,                 /* flush */
        NULL, //socket_vol_file_get,                   /* get */
        NULL, //socket_vol_file_misc,                  /* misc */
        NULL, //socket_vol_file_optional,              /* optional */
        socket_vol_file_close                  /* close */
    },
    {                                          /* group_cls */
        socket_vol_group_create,                       /* create */
        socket_vol_group_open,                         /* open */
        NULL, //socket_vol_group_get,                  /* get */
        socket_vol_group_close                 /* close */
    },
    {                                          /* link_cls */
        NULL, //socket_vol_link_create,                /* create */
        NULL, //socket_vol_link_move,                  /* move */
        NULL, //socket_vol_link_iterate,               /* iterate */
        NULL, //socket_vol_link_get,                   /* get */
        NULL, //socket_vol_link_remove                 /* remove */
    },
    {                                          /* object_cls */
        NULL, //socket_vol_object_open,                /* open */
        NULL, //socket_vol_object_copy,                /* copy */
        NULL, //socket_vol_object_visit,               /* visit */
        NULL, //socket_vol_object_get,                 /* get */
        NULL, //socket_vol_object_misc,                /* misc */
        NULL, //socket_vol_object_optional,            /* optional */
        NULL, //socket_vol_object_close                /* close */
    }
};


static int
socket_vol_init(void)
{
    if((socket_vol_id_g = H5VLregister(&socket_vol_g)) < 0)
        ERROR("H5VLregister failed");
    socket_vol_init_g = 1;

    return(SUCCEED);

error:
    return(FAIL);
} /* end socket_vol_init() */


int
hdf5_socket_vol_set(hid_t fapl_id, char *server_addr, uint16_t server_port)
{
    socket_vol_info_t vol_info = {NULL};

    if(!socket_vol_init_g)
        if(socket_vol_init() < 0)
            ERROR("socket_vol_init failed");

    if(NULL == (vol_info.server_info = (socket_vol_server_info_t *)calloc(1, sizeof(socket_vol_server_info_t))))
        ERROR("calloc of server info failed");

    vol_info.server_info->server_sockaddr.sin_family = AF_INET;
    vol_info.server_info->server_sockaddr.sin_addr.s_addr = inet_addr(server_addr);
    vol_info.server_info->server_sockaddr.sin_port = htons(server_port);

    if(H5Pset_vol(fapl_id, socket_vol_id_g, &vol_info) < 0)
        ERROR("H5Pset_vol failed");

    vol_info.server_info->fapl_ref = 1;
    vol_info.server_info->obj_ref = 0;

    return(SUCCEED);

error:
    if(vol_info.server_info)
        free(vol_info.server_info);

    return(FAIL);
} /* end socket_vol_set() */


static int
socket_vol_connect(socket_vol_server_info_t *server_info)
{
    assert(server_info->obj_ref == 0);

    if((server_info->sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
        ERROR("socket failed");

    if(connect(server_info->sock, (struct sockaddr *)&(server_info->server_sockaddr), sizeof(server_info->server_sockaddr)) < 0)
        ERROR("connect failed");

    return(server_info->sock);

error:
    return(FAIL);
} /* end socket_vol_connect() */


static int
socket_vol_obj_free(socket_vol_t *obj)
{
    assert(obj->server_info->obj_ref > 0);

    /* Decrement reference on server */
    obj->server_info->obj_ref--;
    if(obj->server_info->obj_ref == 0) {
        /* We closed the last object holding this connection open, close */
        if(close(obj->server_info->sock) < 0)
            ERROR("close failed");

        /* If this is also the last item holding the server_info struct open,
         * free it */
        if(obj->server_info->fapl_ref == 0)
            free(obj->server_info);
    } /* end if */

    /* Free object */
    free(obj);

    return(SUCCEED);

error:
    return(FAIL);
} /* end socket_vol_obj_free() */


static void *
socket_vol_fapl_copy(const void *_old_info)
{
    const socket_vol_info_t *old_info = (const socket_vol_info_t *)_old_info;
    socket_vol_info_t *new_info = NULL;

    if(NULL == (new_info = (socket_vol_info_t *)malloc(sizeof(socket_vol_info_t))))
        ERROR("malloc of vol info failed");
    memcpy(new_info, old_info, sizeof(socket_vol_info_t));

    new_info->server_info->fapl_ref++;

    return(new_info);

error:
    if(new_info)
        free(new_info);

    return(NULL);
} /* end socket_vol_fapl_copy() */


static herr_t
socket_vol_fapl_free(void *_info)
{
    socket_vol_info_t *info = (socket_vol_info_t *)_info;

    info->server_info->fapl_ref--;

    if(info->server_info->fapl_ref == 0 && info->server_info->obj_ref == 0) {
        free(info->server_info);
    } /* end if */

    free(info);

    return(SUCCEED);
} /* end socket_vol_fapl_free() */


static void *
socket_vol_file_create(const char *name, unsigned flags, hid_t fcpl_id,
    hid_t fapl_id, hid_t req)
{
    unsigned char *buf = NULL;
    int64_t rbuf;
    unsigned char *p = NULL;
    size_t tmp_size;
    size_t name_len;
    uint64_t msg_len;
    uint64_t fcpl_len;
    uint64_t fapl_len;
    size_t tot_len;
    uint16_t msg_id = SOCKET_VOL_FILE_CREATE_ID;
    const socket_vol_info_t *vol_info;
    socket_vol_t *file = NULL;

    /* Retrieve vol info and allocate file object */
    if(NULL == (vol_info = (socket_vol_info_t *)H5Pget_vol_info(fapl_id)))
        ERROR("H5Pget_vol_info failed");
    if(NULL == (file = (socket_vol_t *)malloc(sizeof(socket_vol_t))))
        ERROR("malloc of file object failed");
    file->server_info = vol_info->server_info;

    /*
     * Calculate message length
     */
    /* Constant size field */
    msg_len = 4;                /* flags */

    /* Calculate fcpl length (and length of fcpl length) */
    if(H5Pencode(fcpl_id, NULL, &tmp_size) < 0)
        ERROR("H5Pencode for size discovery failed");
    fcpl_len = (uint64_t)tmp_size;
    msg_len += fcpl_len + 8;

    /* Calculate fapl length (and length of fapl length) */
    if(H5Pencode(fapl_id, NULL, &tmp_size) < 0)
        ERROR("H5Pencode for size discovery failed");
    fapl_len = (uint64_t)tmp_size;
    msg_len += fapl_len + 8;

    /* Calculate name length */
    name_len = strlen(name) + 1;
    msg_len += name_len;

    /* Calculate total message length */
    tot_len = (size_t)msg_len + SOCKET_VOL_MSG_HEADER_LEN;

    /* Allocate message buffer */
    if(NULL == (buf = (unsigned char *)malloc(tot_len)))
        ERROR("malloc of buffer failed");
    p = buf;

    /*
     * Encode the message
     */
    /* Encode header fields */
    UINT16ENCODE(p, msg_id);
    UINT64ENCODE(p, msg_len);

    /* Encode flags */
    UINT32ENCODE(p, flags);

    /* Encode fcpl length */
    UINT64ENCODE(p, fcpl_len);

    /* Encode fcpl */
    tmp_size = fcpl_len;
    if(H5Pencode(fcpl_id, p, &tmp_size) < 0)
        ERROR("H5Pencode failed");
    p += tmp_size;

    /* Encode fapl length */
    UINT64ENCODE(p, fapl_len);

    /* Encode fapl */
    tmp_size = fapl_len;
    if(H5Pencode(fapl_id, p, &tmp_size) < 0)
        ERROR("H5Pencode failed");
    p += tmp_size;

    /* Encode name */
    strncpy((char *)p, name, name_len);
    p += name_len;
    assert((char)p[-1] == '\0');

    /*
     * Connect to the server, if necessary
     */
    if(vol_info->server_info->obj_ref == 0)
        if(socket_vol_connect(vol_info->server_info) < 0)
            ERROR("socket_vol_connect failed");

    /*
     * Send the message
     */
    if(send(vol_info->server_info->sock, buf, tot_len, 0) != tot_len)
        ERROR("send of message failed");

    /*
     * Receive the ID
     */
    if(8 != recv(vol_info->server_info->sock, &rbuf, 8, MSG_WAITALL))
        ERROR("recv of file id failed");

    /*
     * Set up file object
     */
    p = (unsigned char *)&rbuf;
    INT64DECODE(p, file->obj_id);
    if(file->obj_id < 0)
        ERROR("server returned error");
    file->server_info->obj_ref++;

    free(buf);
    buf = NULL;

    return(file);

error:
    if(file)
        free(file);
    if(buf)
        free(buf);

    return(NULL);
} /* end socket_vol_file_create() */


static void *
socket_vol_file_open(const char *name, unsigned flags, hid_t fapl_id, hid_t req)
{
    unsigned char *buf = NULL;
    int64_t rbuf;
    unsigned char *p = NULL;
    size_t tmp_size;
    size_t name_len;
    uint64_t msg_len;
    uint64_t fapl_len;
    size_t tot_len;
    uint16_t msg_id = SOCKET_VOL_FILE_OPEN_ID;
    const socket_vol_info_t *vol_info;
    socket_vol_t *file = NULL;

    /* Retrieve vol info and allocate file object */
    if(NULL == (vol_info = (socket_vol_info_t *)H5Pget_vol_info(fapl_id)))
        ERROR("H5Pget_vol_info failed");
    if(NULL == (file = (socket_vol_t *)malloc(sizeof(socket_vol_t))))
        ERROR("malloc of file object failed");
    file->server_info = vol_info->server_info;

    /*
     * Calculate message length
     */
    /* Constant size field */
    msg_len = 4;                /* flags */

    /* Calculate fapl length (and length of fapl length) */
    if(H5Pencode(fapl_id, NULL, &tmp_size) < 0)
        ERROR("H5Pencode for size discovery failed");
    fapl_len = (uint64_t)tmp_size;
    msg_len += fapl_len + 8;

    /* Calculate name length */
    name_len = strlen(name) + 1;
    msg_len += name_len;

    /* Calculate total message length */
    tot_len = (size_t)msg_len + SOCKET_VOL_MSG_HEADER_LEN;

    /* Allocate message buffer */
    if(NULL == (buf = (unsigned char *)malloc(tot_len)))
        ERROR("malloc of buffer failed");
    p = buf;

    /*
     * Encode the message
     */
    /* Encode header fields */
    UINT16ENCODE(p, msg_id);
    UINT64ENCODE(p, msg_len);

    /* Encode flags */
    UINT32ENCODE(p, flags);

    /* Encode fapl length */
    UINT64ENCODE(p, fapl_len);

    /* Encode fapl */
    tmp_size = fapl_len;
    if(H5Pencode(fapl_id, p, &tmp_size) < 0)
        ERROR("H5Pencode failed");
    p += tmp_size;

    /* Encode name */
    strncpy((char *)p, name, name_len);
    p += name_len;
    assert((char)p[-1] == '\0');

    /*
     * Connect to the server, if necessary
     */
    if(vol_info->server_info->obj_ref == 0)
        if(socket_vol_connect(vol_info->server_info) < 0)
            ERROR("socket_vol_connect failed");

    /*
     * Send the message
     */
    if(send(vol_info->server_info->sock, buf, tot_len, 0) != tot_len)
        ERROR("send of message failed");

    /*
     * Receive the ID
     */
    if(8 != recv(vol_info->server_info->sock, &rbuf, 8, MSG_WAITALL))
        ERROR("recv of file id failed");

    /*
     * Set up file object
     */
    p = (unsigned char *)&rbuf;
    INT64DECODE(p, file->obj_id);
    if(file->obj_id < 0)
        ERROR("server returned error");
    file->server_info->obj_ref++;

    free(buf);
    buf = NULL;

    return(file);

error:
    if(file)
        free(file);
    if(buf)
        free(buf);

    return(NULL);
} /* end socket_vol_file_open() */


static herr_t
socket_vol_file_close(void *_file, hid_t req)
{
    socket_vol_t *file = (socket_vol_t *)_file;
    unsigned char *p = NULL;
    const uint64_t msg_len = 8;                         /* obj_id */
    size_t tot_len;
    unsigned char buf[SOCKET_VOL_MSG_HEADER_LEN + 8];
    unsigned char rbuf;
    uint16_t msg_id = SOCKET_VOL_FILE_CLOSE_ID;

    assert(file->server_info);
    assert(file->server_info->obj_ref);

    /*
     * Encode the message
     */
    /* Calculate message length */
    tot_len = (size_t)msg_len + SOCKET_VOL_MSG_HEADER_LEN;

    p = buf;

    /* Encode header fields */
    UINT16ENCODE(p, msg_id);
    UINT64ENCODE(p, msg_len);

    /* Encode file id */
    INT64ENCODE(p, file->obj_id);

    /*
     * Send the message
     */
    if(send(file->server_info->sock, buf, tot_len, 0) != tot_len)
        ERROR("send of message failed");

    /*
     * Receive the confirmation
     */
    if(1 != recv(file->server_info->sock, &rbuf, 1, MSG_WAITALL))
        ERROR("recv of confirmation failed");
    if(rbuf != (unsigned char)0)
        ERROR("server returned error");

    /*
     * Free the file object
     */
    if(socket_vol_obj_free(file) < 0)
        ERROR("socket_vol_obj_free failed");
    file = NULL;

    return(SUCCEED);

error:
    if(socket_vol_obj_free(file) < 0)
        PRINT_ERROR("socket_vol_obj_free failed");

    return(FAIL);
} /* end socket_vol_file_close() */


static void *
socket_vol_group_create(void *_obj, H5VL_loc_params_t loc_params,
    const char *name, hid_t gcpl_id, hid_t gapl_id, hid_t req)
{
    unsigned char *buf = NULL;
    int64_t rbuf;
    socket_vol_t *obj = (socket_vol_t *)_obj;
    unsigned char *p = NULL;
    size_t tmp_size;
    size_t name_len;
    uint64_t msg_len;
    uint64_t lcpl_len;
    uint64_t gcpl_len;
    uint64_t gapl_len;
    size_t tot_len;
    uint16_t msg_id = SOCKET_VOL_GROUP_CREATE_ID;
    hid_t lcpl_id = -1;
    socket_vol_t *grp = NULL;

    assert(loc_params.type == H5VL_OBJECT_BY_SELF);

    /* Allocate group object */
    if(NULL == (grp = (socket_vol_t *)malloc(sizeof(socket_vol_t))))
        ERROR("malloc of group object failed");
    grp->server_info = obj->server_info;

    assert(grp->server_info);
    assert(grp->server_info->obj_ref);

    /*
     * Get lcpl id from gcpl
     */
    if(H5Pget(gcpl_id, H5VL_GRP_LCPL_ID, &lcpl_id) < 0)
        ERROR("H5Pget failed to get lcpl id from gcpl");

    /*
     * Calculate message length
     */
    /* Constant size field */
    msg_len = 8;                /* obj_id */

    /* Calculate lcpl length (and length of lcpl length) */
    if(H5Pencode(lcpl_id, NULL, &tmp_size) < 0)
        ERROR("H5Pencode for size discovery failed");
    lcpl_len = (uint64_t)tmp_size;
    msg_len += lcpl_len + 8;

    /* Calculate gcpl length (and length of gcpl length) */
    if(H5Pencode(gcpl_id, NULL, &tmp_size) < 0)
        ERROR("H5Pencode for size discovery failed");
    gcpl_len = (uint64_t)tmp_size;
    msg_len += gcpl_len + 8;

    /* Calculate gapl length (and length of gapl length) */
    if(H5Pencode(gapl_id, NULL, &tmp_size) < 0)
        ERROR("H5Pencode for size discovery failed");
    gapl_len = (uint64_t)tmp_size;
    msg_len += gapl_len + 8;

    /* Calculate name length */
    name_len = strlen(name) + 1;
    msg_len += name_len;

    /* Calculate total message length */
    tot_len = (size_t)msg_len + SOCKET_VOL_MSG_HEADER_LEN;

    /* Allocate message buffer */
    if(NULL == (buf = (unsigned char *)malloc(tot_len)))
        ERROR("malloc of buffer failed");
    p = buf;

    /*
     * Encode the message
     */
    /* Encode header fields */
    UINT16ENCODE(p, msg_id);
    UINT64ENCODE(p, msg_len);

    /* Encode object id */
    INT64ENCODE(p, obj->obj_id);

    /* Encode lcpl length */
    UINT64ENCODE(p, lcpl_len);

    /* Encode lcpl */
    tmp_size = lcpl_len;
    if(H5Pencode(lcpl_id, p, &tmp_size) < 0)
        ERROR("H5Pencode failed");
    p += tmp_size;

    /* Encode gcpl length */
    UINT64ENCODE(p, gcpl_len);

    /* Encode gcpl */
    tmp_size = gcpl_len;
    if(H5Pencode(gcpl_id, p, &tmp_size) < 0)
        ERROR("H5Pencode failed");
    p += tmp_size;

    /* Encode gapl length */
    UINT64ENCODE(p, gapl_len);

    /* Encode gapl */
    tmp_size = gapl_len;
    if(H5Pencode(gapl_id, p, &tmp_size) < 0)
        ERROR("H5Pencode failed");
    p += tmp_size;

    /* Encode name */
    strncpy((char *)p, name, name_len);
    p += name_len;
    assert((char)p[-1] == '\0');

    /*
     * Send the message
     */
    if(send(grp->server_info->sock, buf, tot_len, 0) != tot_len)
        ERROR("send of message failed");

    /*
     * Receive the ID
     */
    if(8 != recv(grp->server_info->sock, &rbuf, 8, MSG_WAITALL))
        ERROR("recv of group id failed");

    /*
     * Set up group object
     */
    p = (unsigned char *)&rbuf;
    INT64DECODE(p, grp->obj_id);
    if(grp->obj_id < 0)
        ERROR("server returned error");
    grp->server_info->obj_ref++;

    free(buf);
    buf = NULL;

    return(grp);

error:
    if(grp)
        free(grp);
    if(buf)
        free(buf);

    return(NULL);
} /* end socket_vol_group_create() */


static void *
socket_vol_group_open(void *_obj, H5VL_loc_params_t loc_params,
    const char *name, hid_t gapl_id, hid_t req)
{
    unsigned char *buf = NULL;
    int64_t rbuf;
    socket_vol_t *obj = (socket_vol_t *)_obj;
    unsigned char *p = NULL;
    size_t tmp_size;
    size_t name_len;
    uint64_t msg_len;
    uint64_t gapl_len;
    size_t tot_len;
    uint16_t msg_id = SOCKET_VOL_GROUP_OPEN_ID;
    socket_vol_t *grp = NULL;

    assert(loc_params.type == H5VL_OBJECT_BY_SELF);

    /* Allocate group object */
    if(NULL == (grp = (socket_vol_t *)malloc(sizeof(socket_vol_t))))
        ERROR("malloc of group id failed");
    grp->server_info = obj->server_info;

    assert(grp->server_info);
    assert(grp->server_info->obj_ref);

    /*
     * Calculate message length
     */
    /* Constant size field */
    msg_len = 8;                /* obj_id */

    /* Calculate gapl length (and length of gapl length) */
    if(H5Pencode(gapl_id, NULL, &tmp_size) < 0)
        ERROR("H5Pencode for size discovery failed");
    gapl_len = (uint64_t)tmp_size;
    msg_len += gapl_len + 8;

    /* Calculate name length */
    name_len = strlen(name) + 1;
    msg_len += name_len;

    /* Calculate total message length */
    tot_len = (size_t)msg_len + SOCKET_VOL_MSG_HEADER_LEN;

    /* Allocate message buffer */
    if(NULL == (buf = (unsigned char *)malloc(tot_len)))
        ERROR("malloc of buffer failed");
    p = buf;

    /*
     * Encode the message
     */
    /* Encode header fields */
    UINT16ENCODE(p, msg_id);
    UINT64ENCODE(p, msg_len);

    /* Encode object id */
    INT64ENCODE(p, obj->obj_id);

    /* Encode gapl length */
    UINT64ENCODE(p, gapl_len);

    /* Encode gapl */
    tmp_size = gapl_len;
    if(H5Pencode(gapl_id, p, &tmp_size) < 0)
        ERROR("H5Pencode failed");
    p += tmp_size;

    /* Encode name */
    strncpy((char *)p, name, name_len);
    p += name_len;
    assert((char)p[-1] == '\0');

    /*
     * Send the message
     */
    if(send(grp->server_info->sock, buf, tot_len, 0) != tot_len)
        ERROR("send of message failed");

    /*
     * Receive the ID
     */
    if(8 != recv(grp->server_info->sock, &rbuf, 8, MSG_WAITALL))
        ERROR("recv of group id failed");

    /*
     * Set up group object
     */
    p = (unsigned char *)&rbuf;
    INT64DECODE(p, grp->obj_id);
    if(grp->obj_id < 0)
        ERROR("server returned error");
    grp->server_info->obj_ref++;

    free(buf);
    buf = NULL;

    return(grp);

error:
    if(grp)
        free(grp);
    if(buf)
        free(buf);

    return(NULL);
} /* end socket_vol_group_open() */


static herr_t
socket_vol_group_close(void *_grp, hid_t req)
{
    socket_vol_t *grp = (socket_vol_t *)_grp;
    unsigned char *p = NULL;
    const uint64_t msg_len = 8;                         /* obj_id */
    size_t tot_len;
    unsigned char buf[SOCKET_VOL_MSG_HEADER_LEN + 8];
    unsigned char rbuf;
    uint16_t msg_id = SOCKET_VOL_GROUP_CLOSE_ID;

    assert(grp->server_info);
    assert(grp->server_info->obj_ref);

    /*
     * Encode the message
     */
    /* Calculate message length */
    tot_len = (size_t)msg_len + SOCKET_VOL_MSG_HEADER_LEN;

    p = buf;

    /* Encode header fields */
    UINT16ENCODE(p, msg_id);
    UINT64ENCODE(p, msg_len);

    /* Encode group id */
    INT64ENCODE(p, grp->obj_id);

    /*
     * Send the message
     */
    if(send(grp->server_info->sock, buf, tot_len, 0) != tot_len)
        ERROR("send of message failed");

    /*
     * Receive the confirmation
     */
    if(1 != recv(grp->server_info->sock, &rbuf, 1, MSG_WAITALL))
        ERROR("recv of confirmation failed");
    if(rbuf != (unsigned char)0)
        ERROR("server returned error");

    /*
     * Free the group object
     */
    if(socket_vol_obj_free(grp) < 0)
        ERROR("socket_vol_obj_free failed");
    grp = NULL;

    return(SUCCEED);

error:
    if(socket_vol_obj_free(grp) < 0)
        PRINT_ERROR("socket_vol_obj_free failed");

    return(FAIL);
} /* end socket_vol_group_close() */


static void *
socket_vol_dataset_create(void *_obj, H5VL_loc_params_t loc_params,
    const char *name, hid_t dcpl_id, hid_t dapl_id, hid_t req)
{
    unsigned char *buf = NULL;
    int64_t rbuf;
    socket_vol_t *obj = (socket_vol_t *)_obj;
    unsigned char *p = NULL;
    size_t tmp_size;
    size_t name_len;
    uint64_t msg_len;
    uint64_t dtype_len;
    uint64_t space_len;
    uint64_t lcpl_len;
    uint64_t dcpl_len;
    uint64_t dapl_len;
    size_t tot_len;
    uint16_t msg_id = SOCKET_VOL_DSET_CREATE_ID;
    hid_t dtype_id = -1;
    hid_t space_id = -1;
    hid_t lcpl_id = -1;
    socket_vol_t *dset = NULL;

    assert(loc_params.type == H5VL_OBJECT_BY_SELF);

    /* Allocate dataset object */
    if(NULL == (dset = (socket_vol_t *)malloc(sizeof(socket_vol_t))))
        ERROR("malloc of group object failed");
    dset->server_info = obj->server_info;

    assert(dset->server_info);
    assert(dset->server_info->obj_ref);

    /*
     * Get dtype, space, and lcpl ids from dcpl
     */
    if(H5Pget(dcpl_id, H5VL_DSET_TYPE_ID, &dtype_id) < 0)
        ERROR("H5Pget failed to get datatype id from dcpl");
    if(H5Pget(dcpl_id, H5VL_DSET_SPACE_ID, &space_id) < 0)
        ERROR("H5Pget failed to get dataspace id from dcpl");
    if(H5Pget(dcpl_id, H5VL_DSET_LCPL_ID, &lcpl_id) < 0)
        ERROR("H5Pget failed to get lcpc id from dcpl");

    /*
     * Calculate message length
     */
    /* Constant size field */
    msg_len = 8;                /* obj_id */

    /* Calculate dtype length (and length of dtype length) */
    if(H5Tencode(dtype_id, NULL, &tmp_size) < 0)
        ERROR("H5Tencode for size discovery failed");
    dtype_len = (uint64_t)tmp_size;
    msg_len += dtype_len + 8;

    /* Calculate space length (and length of space length) */
    if(H5Sencode(space_id, NULL, &tmp_size) < 0)
        ERROR("H5Sencode for size discovery failed");
    space_len = (uint64_t)tmp_size;
    msg_len += space_len + 8;

    /* Calculate lcpl length (and length of lcpl length) */
    if(H5Pencode(lcpl_id, NULL, &tmp_size) < 0)
        ERROR("H5Pencode for size discovery failed");
    lcpl_len = (uint64_t)tmp_size;
    msg_len += lcpl_len + 8;

    /* Calculate dcpl length (and length of dcpl length) */
    if(H5Pencode(dcpl_id, NULL, &tmp_size) < 0)
        ERROR("H5Pencode for size discovery failed");
    dcpl_len = (uint64_t)tmp_size;
    msg_len += dcpl_len + 8;

    /* Calculate dapl length (and length of dapl length) */
    if(H5Pencode(dapl_id, NULL, &tmp_size) < 0)
        ERROR("H5Pencode for size discovery failed");
    dapl_len = (uint64_t)tmp_size;
    msg_len += dapl_len + 8;

    /* Calculate name length */
    name_len = strlen(name) + 1;
    msg_len += name_len;

    /* Calculate total message length */
    tot_len = (size_t)msg_len + SOCKET_VOL_MSG_HEADER_LEN;

    /* Allocate message buffer */
    if(NULL == (buf = (unsigned char *)malloc(tot_len)))
        ERROR("malloc of buffer failed");
    p = buf;

    /*
     * Encode the message
     */
    /* Encode header fields */
    UINT16ENCODE(p, msg_id);
    UINT64ENCODE(p, msg_len);

    /* Encode object id */
    INT64ENCODE(p, obj->obj_id);

    /* Encode dtype length */
    UINT64ENCODE(p, dtype_len);

    /* Encode dtype */
    tmp_size = dtype_len;
    if(H5Tencode(dtype_id, p, &tmp_size) < 0)
        ERROR("H5Tencode failed");
    p += tmp_size;

    /* Encode space length */
    UINT64ENCODE(p, space_len);

    /* Encode space */
    tmp_size = space_len;
    if(H5Sencode(space_id, p, &tmp_size) < 0)
        ERROR("H5Sencode failed");
    p += tmp_size;

    /* Encode lcpl length */
    UINT64ENCODE(p, lcpl_len);

    /* Encode lcpl */
    tmp_size = lcpl_len;
    if(H5Pencode(lcpl_id, p, &tmp_size) < 0)
        ERROR("H5Pencode failed");
    p += tmp_size;

    /* Encode dcpl length */
    UINT64ENCODE(p, dcpl_len);

    /* Encode dcpl */
    tmp_size = dcpl_len;
    if(H5Pencode(dcpl_id, p, &tmp_size) < 0)
        ERROR("H5Pencode failed");
    p += tmp_size;

    /* Encode dapl length */
    UINT64ENCODE(p, dapl_len);

    /* Encode dapl */
    tmp_size = dapl_len;
    if(H5Pencode(dapl_id, p, &tmp_size) < 0)
        ERROR("H5Pencode failed");
    p += tmp_size;

    /* Encode name */
    strncpy((char *)p, name, name_len);
    p += name_len;
    assert((char)p[-1] == '\0');

    /*
     * Send the message
     */
    if(send(dset->server_info->sock, buf, tot_len, 0) != tot_len)
        ERROR("send of message failed");

    /*
     * Receive the ID
     */
    if(8 != recv(dset->server_info->sock, &rbuf, 8, MSG_WAITALL))
        ERROR("recv of dataset id failed");

    /*
     * Set up dataset object
     */
    p = (unsigned char *)&rbuf;
    INT64DECODE(p, dset->obj_id);
    if(dset->obj_id < 0)
        ERROR("server returned error");
    dset->server_info->obj_ref++;

    free(buf);
    buf = NULL;

    return(dset);

error:
    if(dset)
        free(dset);
    if(buf)
        free(buf);

    return(NULL);
} /* end socket_vol_dataset_create() */


static void *
socket_vol_dataset_open(void *_obj, H5VL_loc_params_t loc_params,
    const char *name, hid_t dapl_id, hid_t req)
{
    unsigned char *buf = NULL;
    int64_t rbuf;
    socket_vol_t *obj = (socket_vol_t *)_obj;
    unsigned char *p = NULL;
    size_t tmp_size;
    size_t name_len;
    uint64_t msg_len;
    uint64_t dapl_len;
    size_t tot_len;
    uint16_t msg_id = SOCKET_VOL_DSET_OPEN_ID;
    socket_vol_t *dset = NULL;

    assert(loc_params.type == H5VL_OBJECT_BY_SELF);

    /* Allocate dataset object */
    if(NULL == (dset = (socket_vol_t *)malloc(sizeof(socket_vol_t))))
        ERROR("malloc of group object failed");
    dset->server_info = obj->server_info;

    assert(dset->server_info);
    assert(dset->server_info->obj_ref);

    /*
     * Calculate message length
     */
    /* Constant size field */
    msg_len = 8;                /* obj_id */

    /* Calculate dapl length (and length of dapl length) */
    if(H5Pencode(dapl_id, NULL, &tmp_size) < 0)
        ERROR("H5Pencode for size discovery failed");
    dapl_len = (uint64_t)tmp_size;
    msg_len += dapl_len + 8;

    /* Calculate name length */
    name_len = strlen(name) + 1;
    msg_len += name_len;

    /* Calculate total message length */
    tot_len = (size_t)msg_len + SOCKET_VOL_MSG_HEADER_LEN;

    /* Allocate message buffer */
    if(NULL == (buf = (unsigned char *)malloc(tot_len)))
        ERROR("malloc of buffer failed");
    p = buf;

    /*
     * Encode the message
     */
    /* Encode header fields */
    UINT16ENCODE(p, msg_id);
    UINT64ENCODE(p, msg_len);

    /* Encode object id */
    INT64ENCODE(p, obj->obj_id);

    /* Encode dapl length */
    UINT64ENCODE(p, dapl_len);

    /* Encode dapl */
    tmp_size = dapl_len;
    if(H5Pencode(dapl_id, p, &tmp_size) < 0)
        ERROR("H5Pencode failed");
    p += tmp_size;

    /* Encode name */
    strncpy((char *)p, name, name_len);
    p += name_len;
    assert((char)p[-1] == '\0');

    /*
     * Send the message
     */
    if(send(dset->server_info->sock, buf, tot_len, 0) != tot_len)
        ERROR("send of message failed");

    /*
     * Receive the ID
     */
    if(8 != recv(dset->server_info->sock, &rbuf, 8, MSG_WAITALL))
        ERROR("recv of dataset id failed");

    /*
     * Set up dataset object
     */
    p = (unsigned char *)&rbuf;
    INT64DECODE(p, dset->obj_id);
    if(dset->obj_id < 0)
        ERROR("server returned error");
    dset->server_info->obj_ref++;

    free(buf);
    buf = NULL;

    return(dset);

error:
    if(dset)
        free(dset);
    if(buf)
        free(buf);

    return(NULL);
} /* end socket_vol_dataset_open() */


static herr_t
socket_vol_dataset_read(void *_dset, hid_t mem_type_id, hid_t mem_space_id,
    hid_t file_space_id, hid_t dxpl_id, void *app_buf, hid_t req)
{
    socket_vol_t *dset = (socket_vol_t *)_dset;
    unsigned char *buf = NULL;
    unsigned char rbuf;
    unsigned char *p = NULL;
    size_t tmp_size;
    uint64_t msg_len;
    uint64_t mem_type_len;
    uint64_t mem_space_len;
    uint64_t file_space_len;
    uint64_t dxpl_len;
    size_t app_buf_len;
    size_t tot_len;
    htri_t tri_ret;
    size_t mem_type_elem_size;
    hssize_t mem_extent_npoints;
    H5S_sel_type mem_select_type;
    uint16_t msg_id = SOCKET_VOL_DSET_READ_ID;

    assert(dset->server_info);
    assert(dset->server_info->obj_ref);

    /* Check for unsupported datatypes */
    if((tri_ret = H5Tdetect_class(mem_type_id, H5T_VLEN)) < 0)
        ERROR("H5Tdetect_class failred");
    if(tri_ret)
        ERROR("vlen types not supported");
    if((tri_ret = H5Tdetect_class(mem_type_id, H5T_REFERENCE)) < 0)
        ERROR("H5Tdetect_class failred");
    if(tri_ret)
        ERROR("reference types not supported");

    /*
     * Calculate message length
     */
    /* Constant size field */
    msg_len = 8;                /* dataset id */

    /* Calculate memory type length (and length of mem type length) */
    if(H5Tencode(mem_type_id, NULL, &tmp_size) < 0)
        ERROR("H5Tencode for size discovery failed");
    mem_type_len = (uint64_t)tmp_size;
    msg_len += mem_type_len + 8;

    /* Calculate memory space length (and length of mem space length) */
    if(mem_space_id == H5S_ALL)
        mem_space_len = 0;
    else {
        if(H5Sencode(mem_space_id, NULL, &tmp_size) < 0)
            ERROR("H5Sencode for size discovery failed");
        mem_space_len = (uint64_t)tmp_size;
    } /* end else */
    msg_len += mem_space_len + 8;

    /* Calculate file space length (and length of file space length) */
    if(file_space_id == H5S_ALL)
        file_space_len = 0;
    else {
        if(H5Sencode(file_space_id, NULL, &tmp_size) < 0)
            ERROR("H5Sencode for size discovery failed");
        file_space_len = (uint64_t)tmp_size;
    } /* end else */
    msg_len += file_space_len + 8;

    /* Calculate dxpl length (and length of dxpl length) */
    if(H5Pencode(dxpl_id, NULL, &tmp_size) < 0)
        ERROR("H5Pencode for size discovery failed");
    dxpl_len = (uint64_t)tmp_size;
    msg_len += dxpl_len + 8;

    /* Calculate total message length */
    tot_len = (size_t)msg_len + SOCKET_VOL_MSG_HEADER_LEN;

    /* If both the file space and memory space are set to H5S_ALL, we cannot
     * calculate the size of the app buffer.  In this case, we allow the server
     * to calculate the size and send it to us. */
    if(file_space_id == H5S_ALL && mem_space_id == H5S_ALL)
        mem_select_type = H5S_SEL_ALL;
    else {
        /* Calculate application buffer length */
        if(0 == (mem_type_elem_size = H5Tget_size(mem_type_id)))
            ERROR("H5Tget_size failed");
        if((mem_extent_npoints = H5Sget_simple_extent_npoints(
                mem_space_id == H5S_ALL ? file_space_id : mem_space_id)) < 0)
            ERROR("H5Sget_simple_extent_npoints failed");
        app_buf_len = mem_type_elem_size * (size_t)mem_extent_npoints;

        /* Check if the memory selection is the same as the extent.  If not, we must
         * transmit the buffer to the server so unselected elements are preserved.
         */
        /* Note: This is a temporary solution until H5Dtransfer is implemented */
        if((mem_select_type = H5Sget_select_type(
                mem_space_id == H5S_ALL ? file_space_id : mem_space_id)) < 0)
            ERROR("H5Sget_select_type failed");
        if(mem_select_type != H5S_SEL_ALL)
            msg_len += (uint64_t)app_buf_len;
    } /* end else */

    /* Allocate message buffer */
    if(NULL == (buf = (unsigned char *)malloc(tot_len)))
        ERROR("malloc of buffer failed");
    p = buf;

    /*
     * Encode the message
     */
    /* Encode header fields */
    UINT16ENCODE(p, msg_id);
    UINT64ENCODE(p, msg_len);

    /* Encode dataset id */
    INT64ENCODE(p, dset->obj_id);

    /* Encode memory type length */
    UINT64ENCODE(p, mem_type_len);

    /* Encode memory type */
    tmp_size = mem_type_len;
    if(H5Tencode(mem_type_id, p, &tmp_size) < 0)
        ERROR("H5Tencode failed");
    p += tmp_size;

    /* Encode memory space length */
    UINT64ENCODE(p, mem_space_len);

    /* Encode memory space if it is not H5S_ALL */
    if(mem_space_id != H5S_ALL) {
        tmp_size = mem_space_len;
        if(H5Sencode(mem_space_id, p, &tmp_size) < 0)
            ERROR("H5Sencode failed");
        p += tmp_size;
    } /* end if */

    /* Encode file space length */
    UINT64ENCODE(p, file_space_len);

    /* Encode file space if it is not H5S_ALL */
    if(file_space_id != H5S_ALL) {
        tmp_size = file_space_len;
        if(H5Sencode(file_space_id, p, &tmp_size) < 0)
            ERROR("H5Sencode failed");
        p += tmp_size;
    } /* end if */

    /* Encode dxpl length */
    UINT64ENCODE(p, dxpl_len);

    /* Encode dxpl */
    tmp_size = dxpl_len;
    if(H5Pencode(dxpl_id, p, &tmp_size) < 0)
        ERROR("H5Pencode failed");
    p += tmp_size;

    /*
     * Send the message (except the app buffer)
     */
    if(send(dset->server_info->sock, buf, tot_len, 0) != tot_len)
        ERROR("send of message failed");

    /* Send the app buffer if necessary */
    if(mem_select_type != H5S_SEL_ALL) {
        if(send(dset->server_info->sock, app_buf, app_buf_len, 0) != app_buf_len)
            ERROR("send of data failed");
    } /* end if */
    else if(file_space_id == H5S_ALL && mem_space_id == H5S_ALL) {
        /* Receive the app buffer size */
        assert(tot_len > 8);
        if(8 != recv(dset->server_info->sock, buf, 8, MSG_WAITALL))
            ERROR("recv of app buffer size failed");
        p = buf;
        UINT64DECODE(p, msg_len);
        app_buf_len = (size_t)msg_len;
    } /* end if */

    /*
     * Receive the error status
     */
    if(1 != recv(dset->server_info->sock, &rbuf, 1, MSG_WAITALL))
        ERROR("recv of confirmation failed");
    if(rbuf != (unsigned char)0)
        ERROR("server returned error");

    /*
     * Receive the data into the app buffer
     */
    /* Note - we do not currently pack the buffer if the selection is less than
     * the extent */
    if(app_buf_len != recv(dset->server_info->sock, app_buf, app_buf_len, MSG_WAITALL))
        ERROR("recv of data failed");

    free(buf);
    buf = NULL;

    return(SUCCEED);

error:
    if(buf)
        free(buf);

    return(FAIL);
} /* end socket_vol_dataset_read() */


static herr_t
socket_vol_dataset_write(void *_dset, hid_t mem_type_id, hid_t mem_space_id,
    hid_t file_space_id, hid_t dxpl_id, const void *app_buf, hid_t req)
{
    socket_vol_t *dset = (socket_vol_t *)_dset;
    unsigned char *buf = NULL;
    unsigned char rbuf;
    unsigned char *p = NULL;
    size_t tmp_size;
    uint64_t msg_len;
    uint64_t mem_type_len;
    uint64_t mem_space_len;
    uint64_t file_space_len;
    uint64_t dxpl_len;
    size_t app_buf_len;
    size_t tot_len;
    htri_t tri_ret;
    size_t mem_type_elem_size;
    hssize_t mem_extent_npoints;
    uint16_t msg_id = SOCKET_VOL_DSET_WRITE_ID;

    assert(dset->server_info);
    assert(dset->server_info->obj_ref);

    /* Check for unsupported datatypes */
    if((tri_ret = H5Tdetect_class(mem_type_id, H5T_VLEN)) < 0)
        ERROR("H5Tdetect_class failred");
    if(tri_ret)
        ERROR("vlen types not supported");
    if((tri_ret = H5Tdetect_class(mem_type_id, H5T_REFERENCE)) < 0)
        ERROR("H5Tdetect_class failred");
    if(tri_ret)
        ERROR("reference types not supported");

    /*
     * Calculate message length
     */
    /* Constant size field */
    msg_len = 8;                /* dataset id */

    /* Calculate memory type length (and length of mem type length) */
    if(H5Tencode(mem_type_id, NULL, &tmp_size) < 0)
        ERROR("H5Tencode for size discovery failed");
    mem_type_len = (uint64_t)tmp_size;
    msg_len += mem_type_len + 8;

    /* Calculate memory space length (and length of mem space length) */
    if(mem_space_id == H5S_ALL)
        mem_space_len = 0;
    else {
        if(H5Sencode(mem_space_id, NULL, &tmp_size) < 0)
            ERROR("H5Sencode for size discovery failed");
        mem_space_len = (uint64_t)tmp_size;
    } /* end else */
    msg_len += mem_space_len + 8;

    /* Calculate file space length (and length of file space length) */
    if(file_space_id == H5S_ALL)
        file_space_len = 0;
    else {
        if(H5Sencode(file_space_id, NULL, &tmp_size) < 0)
            ERROR("H5Sencode for size discovery failed");
        file_space_len = (uint64_t)tmp_size;
    } /* end else */
    msg_len += file_space_len + 8;

    /* Calculate dxpl length (and length of dxpl length) */
    if(H5Pencode(dxpl_id, NULL, &tmp_size) < 0)
        ERROR("H5Pencode for size discovery failed");
    dxpl_len = (uint64_t)tmp_size;
    msg_len += dxpl_len + 8;

    /* Calculate total message length (for allocation - we do not include the
     * application buffer in this and instead pass the app buffer directly to
     * send) */
    tot_len = (size_t)msg_len + SOCKET_VOL_MSG_HEADER_LEN;

    /* If both the file space and memory space are set to H5S_ALL, we cannot
     * calculate the size of the app buffer.  In this case, we allow the server
     * to calculate the size and send it to us. */
    if(!(file_space_id == H5S_ALL && mem_space_id == H5S_ALL)) { 
        /* Calculate application buffer length */
        /* Note - we do not currently pack the buffer if the selection is less
         * than the extent */
        if(0 == (mem_type_elem_size = H5Tget_size(mem_type_id)))
            ERROR("H5Tget_size failed");
        if((mem_extent_npoints = H5Sget_simple_extent_npoints(
                mem_space_id == H5S_ALL ? file_space_id : mem_space_id)) < 0)
            ERROR("H5Sget_simple_extent_npoints failed");
        app_buf_len = mem_type_elem_size * (size_t)mem_extent_npoints;
        msg_len += app_buf_len;
    } /* end if */

    /* Allocate message buffer */
    if(NULL == (buf = (unsigned char *)malloc(tot_len)))
        ERROR("malloc of buffer failed");
    p = buf;

    /*
     * Encode the message
     */
    /* Encode header fields */
    UINT16ENCODE(p, msg_id);
    UINT64ENCODE(p, msg_len);

    /* Encode dataset id */
    INT64ENCODE(p, dset->obj_id);

    /* Encode memory type length */
    UINT64ENCODE(p, mem_type_len);

    /* Encode memory type */
    tmp_size = mem_type_len;
    if(H5Tencode(mem_type_id, p, &tmp_size) < 0)
        ERROR("H5Tencode failed");
    p += tmp_size;

    /* Encode memory space length */
    UINT64ENCODE(p, mem_space_len);

    /* Encode memory space if it is not H5S_ALL */
    if(mem_space_id != H5S_ALL) {
        tmp_size = mem_space_len;
        if(H5Sencode(mem_space_id, p, &tmp_size) < 0)
            ERROR("H5Sencode failed");
        p += tmp_size;
    } /* end if */

    /* Encode file space length */
    UINT64ENCODE(p, file_space_len);

    /* Encode file space if it is not H5S_ALL */
    if(file_space_id != H5S_ALL) {
        tmp_size = file_space_len;
        if(H5Sencode(file_space_id, p, &tmp_size) < 0)
            ERROR("H5Sencode failed");
        p += tmp_size;
    } /* end if */

    /* Encode dxpl length */
    UINT64ENCODE(p, dxpl_len);

    /* Encode dxpl */
    tmp_size = dxpl_len;
    if(H5Pencode(dxpl_id, p, &tmp_size) < 0)
        ERROR("H5Pencode failed");
    p += tmp_size;

    /*
     * Send the message (except the app buffer)
     */
    if(send(dset->server_info->sock, buf, tot_len, 0) != tot_len)
        ERROR("send of message failed");

    /* Receive the app buffer size if necessary */
    if(file_space_id == H5S_ALL && mem_space_id == H5S_ALL) {
        assert(tot_len > 8);
        if(8 != recv(dset->server_info->sock, buf, 8, MSG_WAITALL))
            ERROR("recv of app buffer size failed");
        p = buf;
        UINT64DECODE(p, msg_len);
        app_buf_len = (size_t)msg_len;
    } /* end if */

    /* Send the app buffer */
    if(send(dset->server_info->sock, app_buf, app_buf_len, 0) != app_buf_len)
        ERROR("send of data failed");

    /*
     * Receive the confirmation
     */
    if(1 != recv(dset->server_info->sock, &rbuf, 1, MSG_WAITALL))
        ERROR("recv of confirmation failed");
    if(rbuf != (unsigned char)0)
        ERROR("server returned error");

    free(buf);
    buf = NULL;

    return(SUCCEED);

error:
    if(buf)
        free(buf);

    return(FAIL);
} /* end socket_vol_dataset_write() */


static herr_t
socket_vol_dataset_close(void *_dset, hid_t req)
{
    socket_vol_t *dset = (socket_vol_t *)_dset;
    unsigned char *p = NULL;
    const uint64_t msg_len = 8;                         /* obj_id */
    size_t tot_len;
    unsigned char buf[SOCKET_VOL_MSG_HEADER_LEN + 8];
    unsigned char rbuf;
    uint16_t msg_id = SOCKET_VOL_DSET_CLOSE_ID;

    assert(dset->server_info);
    assert(dset->server_info->obj_ref);

    /*
     * Encode the message
     */
    /* Calculate message length */
    tot_len = (size_t)msg_len + SOCKET_VOL_MSG_HEADER_LEN;

    p = buf;

    /* Encode header fields */
    UINT16ENCODE(p, msg_id);
    UINT64ENCODE(p, msg_len);

    /* Encode dataset id */
    INT64ENCODE(p, dset->obj_id);

    /*
     * Send the message
     */
    if(send(dset->server_info->sock, buf, tot_len, 0) != tot_len)
        ERROR("send of message failed");

    /*
     * Receive the confirmation
     */
    if(1 != recv(dset->server_info->sock, &rbuf, 1, MSG_WAITALL))
        ERROR("recv of confirmation failed");
    if(rbuf != (unsigned char)0)
        ERROR("server returned error");

    /*
     * Free the dataset object
     */
    if(socket_vol_obj_free(dset) < 0)
        ERROR("socket_vol_obj_free failed");
    dset = NULL;

    return(SUCCEED);

error:
    if(socket_vol_obj_free(dset) < 0)
        PRINT_ERROR("socket_vol_obj_free failed");

    return(FAIL);
} /* end socket_vol_dataset_close() */

